//////////////////////////////////////////////////////////
// 123456789.
// 4
// 5 The game of Sudoku by George Vanecek, version 2.0.
// 2 Send comments and suggestions to vanecek@yahoo.com
// 3 Enjoy! But just remember that you play better if you
// 6 sleep on occassion. Rules are simple: Fill in the grid 
// 9 so that every row, every column, and every 3x3 box
// 8 contains the digits 1 through 9.
// 745913782.

var keyPressed     = 0; // used by applyKey to solve system.event reporting error.
var correctEntries = 0; // The number of correctly places digits, 0..81.

// Predicate used to see if the game is finished:
function inPlay( ) {
  return correctEntries > 0;
}

// Constants:
// Tile file names are indexed by tile type and light/dark:
var NORM    = 0; // normal
var HIGH    = 1; // hightlight
var LIGHT   = 0;
var DARK    = 1;

// Class to map [r,c] to the image filename.
function TileFileNames( ) {
  this.tiles              = new Array(2);
  this.tiles[NORM]        = new Array(2);
  this.tiles[HIGH]        = new Array(2);
  this.tiles[NORM][LIGHT] = "Resources/NormL.png";
  this.tiles[NORM][DARK]  = "Resources/NormD.png";
  this.tiles[HIGH][LIGHT] = "Resources/HighL.png";
  this.tiles[HIGH][DARK]  = "Resources/HighD.png";
  // t is NORM or HIGH and s is LIGHT or DARK:
  this.get = function( t, s ) { 
    return this.tiles[t][s];
  }
  // t is NORM or HIGH, r is 1..9, and c is 1..9:
  this.get2 = function( t, r, c ) {
    return this.tiles[t][((( r < 4 || r > 6 ) && ( c < 4 || c > 6 )) ||
			  ( r > 3 && r < 7 && c > 3 && c < 7 )) 
			 ? DARK 
			 : LIGHT];
  }
}

var tileFileNames = new TileFileNames( );

// A record structure for a move, that is, a digit entered on the board:
function Move( slot, digit, prevDigit, prevMove ) {
  this.slot      = slot;
  this.digit     = digit;
  this.prevDigit = prevDigit;
  this.prevMove  = prevMove;
}

// A stack of Move records as a linked list:
function MoveStack( ) { // Support for Undo
  this.top = null;
  this.push= function( slot, digit, prevDigit ) {
    this.top = new Move( slot, digit, prevDigit, this.top );
  }
  this.pop = function( ) {
    var m = this.top;
    if( m != null ) {
      this.top = m.prevMove;
      m.prevMove = null;
    }
    return m;
  }
  this.clear = function() { 
    this.top = null; 
  }
}
var moveStack = new MoveStack( );

// A class managing the row of nine hint buttons above the board:
function HintButtons( ) {
  var image     = this.base = new Image( );
  image.src     = "Resources/HintBoard.png";
  image.hOffset = 32;
  image.vOffset = 45;
  image.zOrder  = 1;
  image.width   = 254;
  image.height  = 51;
  image.opacity = 0;
  
  this.hint   = 0;             // 0:none, 1..9:show hint for that number
  this.tiles  = new Array(10); // images of 0:none, 1-9:tiles
  this.digits = new Array(10); // text

  var digit;
  for( var d = 1; d <= 9; d++ ) {
    this.tiles[d]   = image = new Image( );
    image.hOffset   = 26 * d + 17;
    image.vOffset   = 58;
    image.zOrder    = 2;
    image.opacity   = 0;
    image.onMouseUp = 'hintButtons.showHint('+d+');';
    image.height    = image.width = 24;
    image.src       = tileFileNames.get( d == this.hint ? HIGH : NORM, DARK );
    this.digits[d]  = digit = new Text( );
    digit.data      = d == 0 ? " " : String( d );
    digit.font      = "Helvetica Bold";
    digit.hOffset   = 26 * d + 23;
    digit.vOffset   = 77;
    digit.color     = preferences.hintDigitColor.value;
    digit.zOrder    = 3;
    digit.opacity   = 0;
  }
  image = digit = null;
  // Show hint for tiles numbered 1-9 or no hint for 0:
  this.showHint = function ( n ) {
    if( inPlay( )) {
      if( this.hint == n ) {
	this.tiles[this.hint].src = tileFileNames.get( NORM, DARK );
	showHintForDigit( this.hint = 0 );
      } else {
	if( this.hint != 0 )
	  this.tiles[this.hint].src = tileFileNames.get( NORM, DARK );
	this.tiles[n].src = tileFileNames.get( HIGH, LIGHT );	
	showHintForDigit( this.hint = n );
      }
    }
  }
  this.showButtons = function( ) {
    this.base.opacity = 255;
    for( var d = 1; d <= 9; d++ )
      this.digits[d].opacity =
      this.tiles[d].opacity  = 255;
  }
  this.hideButton = function( d ) {
    this.base.opacity = 0;
    this.tiles[d].src = tileFileNames.get( NORM, DARK );
    this.digits[d].opacity =
    this.tiles[d].opacity  = 0;
  }
  this.hideButtons = function( ) {
    for( var d = 1; d <= 9; d++ )
      this.hideButton( d );
  }
}
var hintButtons = new HintButtons( );

// A class keeping track of which slot on the board has the focus:
function MouseInTile( ) {
  // (r,c) keeps track of what tile the mouse is in:
  this.r = 0;
  this.c = 0;
  this.set = function( r, c ) {
    if( inPlay( )) {
      this.r = r;
      this.c = c;
    }
  }
  this.isInTile = function( ) {
    return inPlay( ) && r > 0 && c > 0;
  }

  // (clickedRow,clickedCol) keeps track of what tile was last clicked:
  this.clickedRow = 0;
  this.clickedCol = 0;
  this.tileIsChosen = function( ) {
    return inPlay( ) && this.clickedRow > 0 && this.clickedCol > 0;
  }
  this.tileShade = NORM;
  this.tileMode  = LIGHT;
  this.tileClicked = function( ) {
    this.clickedRow = this.r;
    this.clickedCol = this.c;
    this.tileShade  = board[this.r][this.c].nOrH;
    tileFlipTimer.ticking = true;
  }
  // Called by timer when tileFlipTimer.ticking is true:
  this.flipClickedTile = function( ) {
    var slot = board[this.clickedRow][this.clickedCol];
    slot.tile.src =
      tileFileNames.tiles[ slot.nOrH ][this.tileMode ];
    this.tileMode = this.tileMode == LIGHT ? DARK : LIGHT;
  }
}

var mouseInTile = new MouseInTile( );

// A container for all the stuff for a given slot on a board:
function BoardSlot( row, column ) {
  function createTile( ) {
    var tile     = new Image( );
    tile.src     = tileFileNames.get2( NORM, row, column );
    tile.width   = tile.height = 24;
    tile.vOffset = 72 + 26 * row;
    tile.hOffset = 26 * column + 17;
    tile.zOrder  = 1;
    tile.onMouseEnter = 'mouseInTile.set('+row+','+column+');';
    tile.onMouseExit  = 'mouseInTile.set( 0, 0 );';
    return tile;
  };
  function createText( ) {
    var text     = new Text( );
    text.data    = " ";
    text.font    = "Helvetica Bold";
    text.opacity = 0;
    text.width   = text.height = 24;
    text.vOffset = 91 + 26 * row;
    text.hOffset = 26 * column + 23;
    text.color   = preferences.originalDigitColor.value;
    text.zOrder  = 2;
    return text;
  };
  this.r      = row;
  this.c      = column;
  this.tile   = createTile( );
  this.text   = createText( );
  this.shown  = 0; // digit version of char in text.data[0]
  this.nOrH   = NORM;
  this.answer = 0;
}

// Create board as a 10x10 matrix with 1..9, 1..9 set of board slots.
// Row 0 and col 0 are not used:
function createBoard( ) {
  var board = new Array( 10 );
  for( var r = 1; r <= 9; r++ ) {
    board[r] = new Array(10);
    for( var c = 1; c <= 9; c++ )
      board[r][c] = new BoardSlot( r, c );
  }
  return board;
}
var board = createBoard( );

var sn = new Array( 0, 1, 1, 1, 4, 4, 4, 7, 7, 7 );

function isCandidateForDigit( r, c, digit ) {
  if( digit == 0 )
    return true;
  if( board[r][c].shown > 0 )
    return false;
  // Check the row and column:
  for( var i = 1; i <= 9; i++ )
    if( board[r][i].shown == digit || board[i][c].shown == digit )
      return false;
  // Check the 3x3 quadrant:
  for( var r1 = 0; r1 < 3; r1++ ) {
    var ri = sn[ r ] + r1;
    for( var c1 = 0; c1 < 3; c1++ )
      if( board[ ri ][ sn[ c ] + c1 ].shown == digit )
        return false;
  }
  return true;
}

// Set of digits {1..9} used by updateSlotsTooltipHints function:
function DigitSet( ) {
  this.ds    = new Array( 9 );
  this.dsn   = 0;
  this.add   = function( d ) { this.ds[ this.dsn++ ] = d; }
  this.clear = function( )   { this.dsn = 0; }
  this.get   = function( i ) { return this.ds[ i ]; }
}
var ds = new DigitSet( );

// Add a tooltip to each empty tile with the list of possible digits
// such as "Try 1, 2, or 7":
function updateSlotsTooltipHints( ) {
  var ttHint;
  for( r = 1; r <= 9; r++ )
    for( c = 1; c <= 9; c++ ) {
      ds.clear( );
      for( d = 1; d <= 9; d++ )
	if( isCandidateForDigit( r, c, d ))
          ds.add( d );
      if( ds.dsn > 0 ) {
        ttHint = "Try " + ds.get( 0 );
        for( i = 1; i < ds.dsn-1; i++ )
          ttHint = ttHint + ", " + ds.get( i );
        if( ds.dsn > 1 )
          ttHint = ttHint + " or " + ds.get( ds.dsn-1 );
        board[r][c].tile.tooltip = ttHint;
      }
    }
}

// digit 0: for none, 1-9 for that digit:
function showHintForDigit( digit ) {
  var slot;
  for( var r = 1; r <= 9; r++ )
    for( var c = 1; c <= 9; c++ ) {
      slot          = board[r][c];
      slot.nOrH     = isCandidateForDigit( r, c, digit ) ? NORM : HIGH;
      slot.tile.src = 
	tileFileNames.get2( slot.nOrH, r, c );
    }
}

// Create a new game and update the board.
// nHints is the number of hints, the number of digits to show in the solution.
// this is controlled by the easy, normal or hard property values:
function newGame( nHints ) {
  var fn = new Array( 0, 1, 4, 7, 2, 5, 8, 3, 6, 9 );
  function randDigit( n ) {
    return Math.floor( Math.random() * n + 1 );
  }
  // Create default answer matrix:
  var slot;
  for( var r = 1; r <= 9; r++ ) {
    slot            = board[r][1];
    slot.shown      =
    slot.answer     = fn[ r ];
    slot.text.color = preferences.originalDigitColor.value;
    for( var c = 2; c <= 9; c++ ) {
      slot            = board[r][c];
      slot.text.color = preferences.originalDigitColor.value;
      slot.shown      =
      slot.answer     = board[ r ][ c - 1 ].answer % 9 + 1;
    }
  }
  // Randomize the answer matrix:
  for( var d1 = 1; d1 < 9; d1++ ) {
    var d2 = d1;
    while( d1 == d2 )
      d2 = d1 + randDigit( 9-d1 );
    // Change all d1's to d2's and vice versa:
    for( var r = 1; r <= 9; r++ )
      for( var c = 1; c <= 9; c++ ) {
        slot = board[r][c];
	if( slot.answer == d1 )
	  slot.answer = slot.shown = d2;
	else if( slot.answer == d2 )
	  slot.answer = slot.shown = d1;
      }
  }
  // Swap a two rows in each of the three row blocks:
  for( var b = 0; b < 3; b++ ) {
    var r1 = Math.floor( Math.random() * 3 );
    var r2 = r1;
    while( r1 == r2 )
      r2 = Math.floor( Math.random() * 3 );
    r1 = 1 + r1 + b * 3;
    r2 = 1 + r2 + b * 3;
    for( var c = 1; c <= 9; c++ ) {
      var s1 = board[r1][c];
      var s2 = board[r2][c];
      var n1 = s1.answer;
      var n2 = s2.answer;
      s1.answer = s1.shown = n2;
      s2.answer = s2.shown = n1;
    }
  }
  // Swap two columns in each of the three column blocks:
  for( var b = 0; b < 3; b++ ) {
    var c1 = Math.floor( Math.random() * 3 );
    var c2 = c1;
    while( c1 == c2 )
      c2 = Math.floor( Math.random() * 3 );
    c1 = 1 + c1 + b * 3;
    c2 = 1 + c2 + b * 3;
    for( var r = 1; r <= 9; r++ ) {
      var s1 = board[r][c1];
      var s2 = board[r][c2];
      var n1 = s1.answer;
      var n2 = s2.answer;
      s1.answer = s1.shown = n2;
      s2.answer = s2.shown = n1;
    }
  }
  var nToHide = 81 - nHints;
  while( nToHide > 0 ) {
    slot = board[ randDigit( 9 ) ][ randDigit( 9 ) ];
    if( slot.shown > 0 ) {
      slot.shown      = 0;
      slot.text.color = preferences.userDigitColor.value;
      nToHide--;
    }
  }
  for( var r = 1; r <= 9; r++ ) {
    for( var c = 1; c <= 9; c++ ) {
      slot = board[r][c];
      if( slot.shown > 0 ) {
        slot.shown        = board[r][c].answer;
        slot.text.data    = String( slot.shown );
        slot.text.opacity = 255;
      } else {
        slot.shown        = 0;
        slot.text.data    = " ";
        slot.text.opacity = 0;
      }
    }
  }
  tileFlipTimer.ticking = false;
  correctEntries = nHints;
  hintButtons.showButtons( );
  for( var d = 1; d <= 9; d++ )
    checkDigitForCompletion( d );
  updateSlotsTooltipHints( );
}

function easyGame() { 
  newGame( preferences.easyGame.value ); 
}

function normalGame() { 
  newGame( preferences.normalGame.value );
}

function hardGame() { 
  newGame( preferences.hardGame.value ); 
}

// Solve the game and show the answer on the board:
function solve( ) {
  tileFlipTimer.ticking = false;
  hintButtons.hideButtons( );
  showHintForDigit( 0 );
  moveStack.clear();
  var slot;
  for( var r = 1; r <= 9; r++ )
    for( var c = 1; c <= 9; c++ ) {
      slot          = board[r][c];
      slot.nOrH     = NORM;
      slot.tile.src = tileFileNames.get2( slot.nOrH, r, c );
      if( slot.answer != slot.shown )
	slot.text.color = preferences.errorDigitColor.value;
      slot.text.data    = String( slot.answer );
      slot.text.opacity = 255;
    }
  correctEntries = 81;
}

// Called when properties change the text colors:
function applyDigitColors( userColor, originalColor ) {
  for( var r = 1; r <= 9; r++ )
    for( var c = 1; c <= 9; c++ )
      board[r][c].text.color = board[r][c].shown > 0
			     ? originalColor : userColor;
}

function checkDigitForCompletion( d ) {
  var nCorrect = 0;
  for( r = 1; r <= 9; r++ )
    for( c = 1; c <= 9; c++ )
      if( board[r][c].shown == d &&
	  board[r][c].answer == d )
	nCorrect++;
  if( nCorrect == 9 ) {
    hintButtons.hideButton( d );
    return true;
  }
  return false;
}

// Process a key given that its one of 1..9, SPACE, u, m, s, q, e, n, or h:
function applyKey( key ) {    
  if( mouseInTile.tileIsChosen( ) && 
      ((key >= '1' && key <= '9') || key == ' ' )) {
    if( !inPlay( ))
      return;
    var d    = key == ' ' ? 0 : key - '0';
    var slot = board[mouseInTile.clickedRow][mouseInTile.clickedCol];
    if( slot.shown == d )
      return;
    moveStack.push( slot, d, slot.shown );
    if( slot.shown != slot.answer ) {
      slot.text.data    = String( key );
      slot.text.color   = preferences.userDigitColor.value;
      slot.text.opacity = 255;
      slot.shown        = d
      if( d == slot.answer ) {
        // hide hint button is digit is finished.
        if( checkDigitForCompletion( d )) 
          // and unhighlight the board
          showHintForDigit( hintButtons.hint = 0 );
        if( ++correctEntries == 81 ) {
          solve( );
          alert( "You finished!" );        
          return;
        }
      }
    }
    if( hintButtons.hint > 0 )
      showHintForDigit( hintButtons.hint );
    updateSlotsTooltipHints( );
  } else if( key == 'u' ) {
    if( inPlay( ))
      undo( );
  } else if( key == 'm' ) {
    if( inPlay( ))
      showMistakes( );
  } else if( key == 's' ) {
    if( inPlay( ))
      solve( );
  } else if( key == 'q' ) {
    closeWidget( );
  } else if( key == 'e' )
    easyGame( );
  else if( key == 'n' )
    normalGame( );
  else if( key == 'h' )
    hardGame( );  
  system.event.key = undefined;
}

// Undo the last move on the board:
function undo( ) {
  var move = moveStack.pop();
  if( move != null ) {
    move.slot.shown = move.prevDigit;
    move.slot.text.data = move.prevDigit == 0 
                          ? ' ' : String( move.prevDigit );
    mouseInTile.set( move.slot.r, move.slot.c );
    mouseInTile.tileClicked( );
  }
}

// Highlight all wrongly entered digits in red:
function showMistakes( ) {
  tileFlipTimer.ticking = false;
  var slot;
  for( r = 1; r <= 9; r++ )
    for( c = 1; c <= 9; c++ ) {
      slot = board[r][c];
      if( slot.shown != 0 && slot.shown != slot.answer )
        slot.text.color = preferences.errorDigitColor.value;
    }
}

// Called when a mouse button is clicked:
function applyClick( x, y ) {
  if( inPlay( ) && mouseInTile.isInTile())
    mouseInTile.tileClicked( );
  
}

// Display the board:
mainWin.visible = true;
